# -*- coding: utf-8 -*-
import itertools

from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models, connection


class Migration(SchemaMigration):

    def forwards(self, orm):
        self.if_exists_drop_index(u'data_logging', ['object_snapshot'])
        self.if_exists_drop_index(u'data_logging', ['context_data'])

        years = ['2014', '2015', '2016']
        months = [str(x).zfill(2) for x in range(1, 13)]

        for year, month in itertools.product(years, months):
            self.if_exists_drop_index(u'data_logging_y%sm%s' % (year, month), ['object_snapshot'], suffix='_idx')
            self.if_exists_drop_index(u'data_logging_y%sm%s' % (year, month), ['context_data'], suffix='_idx')

    def backwards(self, orm):
        pass

    def if_exists_drop_index(self, table_name, column_names, suffix=''):
        cursor = db._get_connection().cursor()
        tables = connection.introspection.get_table_list(cursor)
        if table_name in tables:
            index_name = ''.join(column_names)
            indexes = connection.introspection.get_indexes(cursor, table_name)
            if index_name in indexes:
                name = db.create_index_name(table_name, column_names, suffix=suffix)
                sql = db.drop_index_string % {
                    "index_name": db.quote_name(name),
                    "table_name": db.quote_name(table_name),
                }
                db.execute(sql)

    models = {
        'datalogging.datalog': {
            'Meta': {'object_name': 'DataLog', 'db_table': "u'data_logging'"},
            'context_data': ('json_field.fields.JSONField', [], {'default': "u'null'", 'null': 'True'}),
            'event_code': ('django.db.models.fields.CharField', [], {'max_length': '32', 'db_index': 'True'}),
            'event_type': ('django.db.models.fields.CharField', [], {'max_length': '256', 'db_index': 'True'}),
            'guid': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'ip': ('django.db.models.fields.CharField', [], {'max_length': '15', 'db_index': 'True'}),
            'object_id': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'db_index': 'True'}),
            'object_snapshot': ('json_field.fields.JSONField', [], {'default': "'{}'", 'null': 'True'}),
            'object_type': ('django.db.models.fields.CharField', [], {'max_length': '96', 'null': 'True', 'db_index': 'True'}),
            'request_token': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'suid': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'db_index': 'True'}),
            'system_type': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'timestamp': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'verbose': ('django.db.models.fields.TextField', [], {'db_index': 'True'})
        }
    }

    complete_apps = ['datalogging']



# coding: utf-8
# сeated on 18 дек. 2014 г.
# author: Michael Vorotyntsev


from datalogging.models import DataLog
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    u"""
    Используется вместо
    $export DJANGO_SETTINGS_MODULE=mysite.settings
    $architect partition --module path.to.the.model.module
    у которого есть проблемы с импортом settings,
    проще написать:
    $python manage.py datalogging_partition_prepare

    """

    help = u'Architect table partition setup'

    def handle(self, *args, **options):
        empty_record = DataLog()
        empty_record.architect.partition.get_partition().prepare()



# coding: utf-8

import re

from django.conf import settings
from m3.db import BaseEnumerate
import six

__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.com'
__docformat__ = 'restructuredtext'


class ConfigurableEnum(type):
    u""" Метакласс автозагрузки типов перечислений.

    Автозагрузка типов осуществляется из settings.py проекта,
    причем данные должны являться словарем с кортежами в качестве
    значений. Наименование параметра должно начинаться с приставки
    "DATALOGGER". Пример:

    ::

        DATALOGGER_EVENT_TYPE = {
            'SYSTEM': ('se', u'Системное событие'),

        }

    После инициализации класса, можно вызывать как обычный аттрибут.
    И в целом класс используется как обычное перечисление. Пример:

    ::

        print EventType.SYSTEM // 'se'

    """
    def __new__(cls, name, bases, dct):
        dct['values'] = dct.get('values') or {}
        parts = re.findall(r'[A-Z][a-z]+', name)
        settings_name = u'_'.join(parts).upper()
        settings_name = u'DATALOGGER_{0}'.format(settings_name)
        constants = getattr(settings, settings_name, {})
        assert isinstance(constants, dict)
        for const_name, value in constants.items():
            assert const_name.isupper()
            assert isinstance(value, tuple) and len(value) == 2
            dct[const_name] = value[0]
            dct['values'].update([value])
        return type.__new__(cls, name, bases, dct)


@six.add_metaclass(ConfigurableEnum)
class EventType(BaseEnumerate):
    u""" Перечисление типов событий.

    Пример возможных типов:
    - (se, Системное событие)
    - (lse, Юридически важное событие)

    """


@six.add_metaclass(ConfigurableEnum)
class EventCode(BaseEnumerate):
    u""" Перечисление кодов событий.

    Пример возможных кодов:
    - (insert, Добавление)
    - (delete, Удаление)

    """

    LOGIN = 'login'
    INSERT = 'insert'
    DELETE = 'delete'
    UPDATE = 'update'
    LOGOUT = 'logout'
    WIN_OPEN = 'win_open'

    values = {
        INSERT: u'Добавление',
        DELETE: u'Удаление',
        UPDATE: u'Изменение',
        LOGIN: u'Пользователь вошел в систему',
        LOGOUT: u'Пользователь вышел из системы',
        WIN_OPEN: u'Пользователь открыл окно'
    }


@six.add_metaclass(ConfigurableEnum)
class SystemsEnum(BaseEnumerate):
    u""" Перечисление сред логирования.

    Пример возможных значений:
    - ('scheduled_task', u'Задача по расписанию')
    - ('ws', u'Web-сервис')

    """



# coding: utf-8
import warnings
from copy import copy

from django.conf import settings
from django.db.models.signals import class_prepared
from django.dispatch import Signal

from .managers import DataLoggingManager, MANAGER_NAME


suspects = set(getattr(settings, 'DATALOGGER_SUSPECTS_MODEL', []))

if not suspects:
    msg = 'Список моделей DATALOGGER_SUSPECTS_MODEL пуст'
    warnings.warn(msg.encode('utf-8'))
    
elif len(suspects) != len(settings.DATALOGGER_SUSPECTS_MODEL):
    # Проверка на дубликаты, в случае их наличия выкидывается warning
    duplicates = list(copy(settings.DATALOGGER_SUSPECTS_MODEL))
    for model_name in suspects:
        duplicates.remove(model_name)
    msg = (u'Дублированное вхождение в DATALOGGER_SUSPECTS_MODEL '
           u'модели(ей):{0}').format(duplicates)
    warnings.warn(msg.encode('utf-8'))


# Сигналы определяющие поведение методов логгера
# =====================================================================

post_snapshot_signal = Signal(providing_args=['log_record'])
post_system_event_signal = Signal(providing_args=['log_record'])
log_context_signal = Signal(providing_args=['request'])
msg_for_log_signal = Signal(
    providing_args=['log_record', 'model_instance', 'fields_dict'])


def add_datalogger_manager(sender, **kw):
    u""" Обработчик сигнала class_prepared.

    Если модель указана в списке ``DATALOGGER_SUSPECTS_MODEL``, то
    обработчик добавляет в класс модели ``DataLoggingManager``.

    """
    cls = sender
    full_path = '{0}.{1}'.format(cls.__module__, cls.__name__)
    short_path = '{0}.{1}'.format(cls._meta.app_label, cls.__name__)
    if suspects.intersection((short_path, full_path)):
        cls.add_to_class(MANAGER_NAME, DataLoggingManager())

if not getattr(settings, 'DATALOGGER_SHUTUP', False):
    class_prepared.connect(add_datalogger_manager, weak=False)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2019-06-28 18:12
from __future__ import unicode_literals

from django.db import migrations, models
import json_field.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DataLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('guid', models.CharField(max_length=64, verbose_name='Уникальный идентификатор записи')),
                ('timestamp', models.DateTimeField(auto_now_add=True, verbose_name='Время совершения действия')),
                ('event_type', models.CharField(db_index=True, max_length=256, verbose_name='Тип события')),
                ('event_code', models.CharField(db_index=True, max_length=32, verbose_name='Код события')),
                ('object_id', models.IntegerField(db_index=True, null=True, verbose_name='ID изменяемого объекта')),
                ('object_type', models.CharField(db_index=True, max_length=96, null=True, verbose_name='Имя модели объекта')),
                ('object_snapshot', json_field.fields.JSONField(default='{}', help_text='Enter a valid JSON object', null=True, verbose_name='Снимок состояния объекта')),
                ('verbose', models.TextField(db_index=True, verbose_name='Человекопонятное описание события')),
                ('system_type', models.CharField(max_length=32, verbose_name='Подсистема')),
                ('suid', models.IntegerField(db_index=True, null=True, verbose_name='Идентификатор объекта в подсистеме')),
                ('ip', models.CharField(db_index=True, max_length=15, verbose_name='IP адрес')),
                ('context_data', json_field.fields.JSONField(default='null', help_text='Enter a valid JSON object', null=True, verbose_name='Контекстно-зависимые данные')),
                ('request_token', models.CharField(max_length=64, verbose_name='Токен запроса')),
            ],
            options={
                'verbose_name': 'Журнал событий',
                'db_table': 'data_logging',
            },
        ),
    ]



# coding: utf-8
"""
.. module:: datalogging.api
.. moduleauthor:: Andrey Torsunov <torsunov@bars-open.ru>

"""
from m3_django_compat import get_model

__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.com'
__docformat__ = 'restructuredtext'

from django.db import models
from django.db.models import Q

from m3_legacy.middleware import get_thread_data
import six


def get_request():
    u"""Получение текущего запроса.
    """
    _thread_locals = get_thread_data()
    try:
        request = _thread_locals.wref_request()
    except AttributeError:
        request = None
    return request


def get_request_token():
    u"""Получение токена текущего запроса.
    """
    _thread_locals = get_thread_data()
    return getattr(_thread_locals, 'request_token', None)


def get_user_ip():
    u""" Получение IP текущего пользователя.

    :return str or None: Если не удалось получить IP вернет None

    """
    ip = None
    try:
        request = get_request()
        ip = request.META.get('X-Real-IP')
        ip = ip or request.META.get('HTTP_X_FORWARDED_FOR')
        ip = ip or request.META.get('REMOTE_ADDR')
        ip = ip.split(',')[-1].strip()[:15]
    finally:
        return ip


def get_events_by_token(request_token):
    u""" Получение записей лога по их общему токену """
    DataLog = get_model('datalogging', 'DataLog')
    return DataLog.objects.filter(
        request_token=request_token
    ).order_by('timestamp')


def filter_events(**kw):
    u""" Фильтрация по записям лолга.

    .. Note :: Не предусмотрена работа с NoSQL.

    Если необходимо произвести фильтрацию по данным находящимся в
    в полях ``context_data`` или ``object_snapshot``. То необходимо
    использовать префиксы ``_context`` и ``_snapshot`` соответственно.

    ::

        filter_events(
            object_type='enterprise.Enterprise',
            system_type=SystemsEnum.MAIN_APPLICATION,
            _context__diff__name='Учреждение №666',
            _context__ent='455',
        )

    """
    DataLog = get_model('datalogging', 'DataLog')
    _filter = {}
    context_filter = Q()
    snapshot_filter = Q()
    prefixes = [
        ('_context', 'context_data', context_filter),
        ('_snapshot', 'object_snapshot', snapshot_filter),
    ]

    for key, value in kw.items():
        for prefix, filter_name, prefix_filter in prefixes:
            if key.startswith(prefix):
                f = key.replace(prefix, '', 1)
                value = six.text_type(value).encode(
                    'unicode-escape').replace('\\u', '\\\\u')
                re = '{0}.*:[^:]+{1}'.format(f.replace('__', '.+'), value)
                filter_name = '{0}__iregex'.format(filter_name)
                q = Q(**{filter_name: re})
                prefix_filter.add(q, Q.AND)
                break
        else:
            _filter[key] = value

    return DataLog.objects.filter(
        context_filter & snapshot_filter, **_filter)



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding index on 'DataLog', fields ['suid']
        db.create_index(u'data_logging', ['suid'])

        # Removing index on 'DataLog', fields ['request_token']
        db.delete_index(u'data_logging', ['request_token'])


    def backwards(self, orm):
        # Adding index on 'DataLog', fields ['request_token']
        db.create_index(u'data_logging', ['request_token'])

        # Removing index on 'DataLog', fields ['suid']
        db.delete_index(u'data_logging', ['suid'])


    models = {
        'datalogging.datalog': {
            'Meta': {'object_name': 'DataLog', 'db_table': "u'data_logging'"},
            'context_data': ('json_field.fields.JSONField', [], {'default': "u'null'", 'null': 'True'}),
            'event_code': ('django.db.models.fields.CharField', [], {'max_length': '32', 'db_index': 'True'}),
            'event_type': ('django.db.models.fields.CharField', [], {'max_length': '256', 'db_index': 'True'}),
            'guid': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'ip': ('django.db.models.fields.CharField', [], {'max_length': '15', 'db_index': 'True'}),
            'object_id': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'db_index': 'True'}),
            'object_snapshot': ('json_field.fields.JSONField', [], {'default': "'{}'", 'null': 'True'}),
            'object_type': ('django.db.models.fields.CharField', [], {'max_length': '96', 'null': 'True', 'db_index': 'True'}),
            'request_token': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'suid': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'db_index': 'True'}),
            'system_type': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'timestamp': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'verbose': ('django.db.models.fields.TextField', [], {'db_index': 'True'})
        }
    }

    complete_apps = ['datalogging']



# coding: utf-8
import uuid
import weakref
from django.utils.deprecation import MiddlewareMixin

from m3_legacy.middleware import get_thread_data

__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.com'
__docformat__ = 'restructuredtext'


class RequestTokenMiddleware(MiddlewareMixin):
    u""" Middleware создающее токен для текущего запроса.

    Токен в последующем устаналвивается во все возникающие события
    в логгере, таким образом можно просмотреть/отследить цепочку
    событий возникших во время запроса.

    """
    def process_request(self, request):
        _thread_local = get_thread_data()
        _thread_local.request_token = str(uuid.uuid4())


class CaptureRequestMiddleware(MiddlewareMixin):
    u""" Middleware сохраняющее текущий request.

    В последующем request используется для получения ряда параметров
    необходимых для логирования действий ( например ip пользователя ).

    """
    def process_request(self, request):
        _thread_local = get_thread_data()
        _thread_local.wref_request = weakref.ref(request)



# coding: utf-8
"""
.. module:: datalogging.management.commands.datalogging_compress
.. moduleauthor:: Rinat Sabitov <sabitov@bars-open.ru>

"""
import os
import sys
import datetime
import tempfile
from zipfile import ZipFile
from django.core.management.base import BaseCommand
from django.conf import settings
from django.core import management


class Command(BaseCommand):
    u"""Класс менежд-команды.
    """
    help = 'Archieve datalogging data'

    def add_arguments(self, parser):
        parser.add_argument(
            '--destination',
            dest='destination',
            default=getattr(settings, 'DATALOGGER_COMPRESS_DESTINATION', ''),
            help='Archieving destination folder')
        parser.add_argument(
            '--reset',
            action='store_true',
            dest='reset',
            default=False,
            help='Delete all data after compress')

    def handle(self, *args, **options):
        u"""Метод обрабатыюващий команду. На входе ожидаются только именованые
        параметры переданные из командной строки, в частности
        `options['reset']` и `options['destination']`.
        """

        default_filename_template = 'datalogging-dump-%d-%m-%Y-%H-%M'
        filename_template = getattr(settings,
                'DATALOGGER_COMPRESS_FILENAME_TEMPLATE',
                default_filename_template)

        with tempfile.NamedTemporaryFile() as tf:
            #так как dumpdata сливает все в stdout, нам не остается ничего
            #другого как подменить сам stdout
            sysout = sys.stdout
            try:
                sys.stdout = tf
                management.call_command('dumpdata', 'datalogging')
            finally:
                sys.stdout = sysout

            filename = datetime.datetime.now().strftime(filename_template)
            with ZipFile(os.path.join(options['destination'],
                '%s.zip' % filename), 'w') as myzip:
                myzip.write(tf.name, filename + '.json')

        if options['reset']:
            management.call_command('reset', 'datalogging')



# coding: utf-8

import architect
import datetime
import uuid
from itertools import chain

from django.db import models
from m3_django_compat import ModelOptions
from django.utils.encoding import force_str
from json_field import JSONField
from json_field.fields import JSONEncoder

from datalogging.api import get_user_ip, get_request_token, get_request
from .helpers import get_snapshot
from .enums import EventCode
from .signals import (
    msg_for_log_signal, log_context_signal, post_snapshot_signal,
    post_system_event_signal)
import six


__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.com'
__docformat__ = 'restructuredtext'


class LogJSONEncoder(JSONEncoder):
    u""" Наследник JSONEncoder, который умеет работать с ugettext. """

    def default(self, obj):
        try:
            ret = super(LogJSONEncoder, self).default(obj)
        except (UnicodeEncodeError, TypeError, ValueError):
            ret = force_str(obj)

        return ret


architect_options = dict(
    db='datalogger',
    type='range',
    subtype='date',
    constraint='month',
    column='timestamp')


@architect.install('partition', **architect_options)
class DataLog(models.Model):
    u""" Лог активности пользователей.

    Лог предусматривает отслеживание активности пользователей в системе,
    такой, как удаление/добавление/изменение записей документов, настро
    ек, прав доступа и т.д.

    `guid`:
        Уникальный ИД записи не является primary_key в случае с
        реляционными БД. Его необходимость возникает в момент выгрузки,
        когда несколько логов выгружаются в одну систему и применение
        primary_key из реляционной БД вызвало бы ряд неудобств связанных
        с коллизиями.
        Во избежание коллизий используется uuid с солью.

    `system_type`:
        В качестве подсистемы выступают сервисы(rest, soap), фоновые
        задачи, основное приложение и т.д. Т.е. это по существу среда
        возникновения события, которое следует перехватить.

    `request_token`:
        В рамках одного запроса сущетствует токен по которому можно
        отследить всю активность пользователя в течении запроса.

    """
    guid = models.CharField(
        verbose_name=u'Уникальный идентификатор записи',
        max_length=64,
        db_index=False)

    timestamp = models.DateTimeField(
        verbose_name=u'Время совершения действия',
        auto_now_add=True,
        db_index=False)

    event_type = models.CharField(
        verbose_name=u'Тип события',
        max_length=256,
        db_index=True)

    event_code = models.CharField(
        verbose_name=u'Код события',
        max_length=32,
        db_index=True)

    object_id = models.IntegerField(
        verbose_name=u'ID изменяемого объекта',
        null=True,
        db_index=True)

    object_type = models.CharField(
        verbose_name=u'Имя модели объекта',
        max_length=96,
        null=True,
        db_index=True)

    object_snapshot = JSONField(
        verbose_name=u'Снимок состояния объекта',
        null=True,
        default='{}',
        encoder=LogJSONEncoder,
        db_index=False)

    verbose = models.TextField(
        verbose_name=u'Человекопонятное описание события',
        db_index=True)

    system_type = models.CharField(
        verbose_name=u'Подсистема',
        max_length=32,
        db_index=False)

    suid = models.IntegerField(
        verbose_name=u'Идентификатор объекта в подсистеме',
        null=True,
        db_index=True)

    ip = models.CharField(
        verbose_name=u'IP адрес',
        max_length=15,
        db_index=True)

    context_data = JSONField(
        verbose_name=u'Контекстно-зависимые данные',
        null=True,
        encoder=LogJSONEncoder,
        db_index=False)

    request_token = models.CharField(
        verbose_name=u'Токен запроса',
        max_length=64,
        db_index=False)

    context_data.encoder_kwargs.update(ensure_ascii=False)
    object_snapshot.encoder_kwargs.update(ensure_ascii=False)

    class Meta:
        verbose_name = u'Журнал событий'
        db_table = u'data_logging'

    def __init__(self, *args, **kwargs):
        u""" Инициализация записи лога.

        В момент инициализации подставляются данные из контекста,
        определенного в конкретном проекте. Для того, чтобы пере
        дать контекст необходимо подписать обработчик на сигнал
        `log_context_signal`, который должен возвращать словарь
        содержащий в качестве ключей `suid`, `system_type`,
        `event_type`.

        """
        super(DataLog, self).__init__(*args, **kwargs)

        # При формировании записи лога не передаются параметры,
        # а при вытаскивании данных из базы наоборот.
        if not (args or kwargs):
            self.context_data = {}
            self.ip = get_user_ip()
            self.guid = uuid.uuid4().hex
            self.request_token = get_request_token()

            request = get_request()
            responses = log_context_signal.send(self, request=request)
            if len(responses) > 0:
                _, log_context = responses[-1]
            else:
                raise NotImplementedError(
                    "`log_context_signal` signal receiver not implemented")

            self.suid = log_context.get('suid')
            self.timestamp = datetime.datetime.now()
            self.event_type = log_context.get('event_type')
            self.system_type = log_context.get('system_type')

    @classmethod
    def _diff_field(cls, instance, attr_name, old_value, name_map):
        u""" Выявление различий значений поля между состояниями модели.

        :param instance: Экземпляр модели
        :param str attr_name: Наименование поля
        :param old_value: Значение поля из предыдущего состояния модели
        :param dict name_map: Поля модели без *_id

        :return tuple: Имя поля, (старое значение, новое значение),
            человекопонятное наименование поля

        """
        alias = attr_name
        if attr_name.endswith('_id') and attr_name[:-3] in name_map:
            alias = attr_name[:-3]
            diff, verbose_name = cls._diff_related_field(
                instance, alias, old_value
            )
        else:
            diff, verbose_name = cls._diff_normal_field(
                instance, alias, old_value
            )

        return alias, diff, verbose_name

    @staticmethod
    def _diff_normal_field(instance, alias, old_value):
        current_value = getattr(instance, alias, None)
        field = ModelOptions(instance).get_field_by_name(alias)[0]
        diff = (old_value or u'(не указано)', current_value)
        return diff, field.verbose_name or alias

    @staticmethod
    def _diff_related_field(instance, alias, old_object_id):
        field = ModelOptions(instance).get_field_by_name(alias)[0]
        model = field.rel.to
        obj = getattr(instance, alias, None)
        new_val, old_val = six.text_type(obj), u'(не указано)'

        if old_object_id:
            try:
                old_val = model.objects.get(id=old_object_id)
            except model.DoesNotExist:
                old_val = u'(не найден)'
            else:
                old_val = six.text_type(old_val)

        return (old_val, new_val), field.verbose_name or alias

    @classmethod
    def get_verbose_msg(cls, log_record, model_instance, **kwargs):
        u""" Получение человекопонятного описания события.

        Т.к. для каждого проекта формирование описания будет разным
        метод работает через сигнал `msg_for_log_signal`.

        """
        fields_dict = model_instance and model_instance.__dict__.copy() or {}
        responses = msg_for_log_signal.send(
            cls,
            log_record=log_record,
            model_instance=model_instance,
            fields_dict=fields_dict,
            **kwargs)

        if len(responses) > 0:
            _, response = responses[0]
        else:
            raise NotImplementedError(
                "`msg_for_log_signal` signal receiver not implemented")

        return response

    @classmethod
    def make_snapshot(cls, instance, event_code):
        u""" Формирование записи лога о состоянии модели.

        Перед сохранением записи лога вызывается сигнал `post_snapshot`,
        где в качестве аргумента передается не сохраненная запись лога,
        что в свою очередь позволяет воздействовать на запись.

        :param instance: Экземпляр модели
        :param event_code: Событие происшедшее с моделью

        """
        log_rec = cls()
        log_rec.object_id = instance.id
        log_rec.object_type = '{0}.{1}'.format(
            instance._meta.app_label,
            instance.__class__.__name__
        )

        log_rec.event_code = event_code

        if event_code != EventCode.DELETE:
            result_diff = {}
            verbose_names = {}
            cur_snapshot = get_snapshot(instance)
            name_map = list(set(chain.from_iterable(
                (field.name, field.attname) if hasattr(field, 'attname') else (
                field.name,)
                for field in instance._meta.get_fields()
                # For complete backwards compatibility, you may want to exclude
                # GenericForeignKey from the results.
                if not (field.many_to_one and field.related_model is None)
            )))

            # Различие в состояниях экземпляра модели
            state_diff = getattr(instance, '_snapshot', set())
            state_diff -= cur_snapshot

            for attr_name, old_value in state_diff:
                alias, field_diff, verbose = cls._diff_field(
                    instance, attr_name, old_value, name_map
                )
                result_diff[alias] = field_diff
                verbose_names[alias] = verbose

            log_rec.context_data['diff'] = result_diff
            log_rec.context_data['field_names'] = verbose_names

        snapshot = {}

        for key, val in instance.__dict__.items():
            if key.startswith('_'):
                continue

            if val is None:
                snapshot[key] = val
            else:
                str_val = force_str(val)
                if str_val.isdigit():
                    snapshot[key] = val
                else:
                    snapshot[key] = str_val

        log_rec.object_snapshot = snapshot
        log_rec.verbose = cls.get_verbose_msg(log_rec, instance)

        # Если хотя бы один обработчик вернет False, то это будет
        # означать, что запись с логом не нужно сохранять.
        signal_results = post_snapshot_signal.send(cls, log_record=log_rec)
        if False not in signal_results:
            log_rec.save()

    @classmethod
    def make_system_event(cls, event_code, context_data=None, **kwargs):
        u""" Формирование записи о событии происшедшем в системе.

        Все спецефичные данные о событии обрабатываются через сигнал
        `post_system_event` ровно так же, как в методе `make_snapshot`.

        :param event_code: Событие в системе.
        :param dict context_data: Доп. данные описывающие событие.

        """
        context_data = context_data or {}
        log_rec = cls()
        log_rec.context_data.update(context_data)
        log_rec.event_code = event_code
        log_rec.verbose = cls.get_verbose_msg(log_rec, None, **kwargs)

        # Если хотя бы один обработчик вернет False, то это будет
        # означать, что запись с логом не нужно сохранять.
        signal_results = post_system_event_signal.send(
            cls, log_record=log_rec, **kwargs)
        if False not in signal_results:
            log_rec.save()

    def __hash__(self):
        return hash(self.guid)


# coding: utf-8

"""
Migrate management command.
"""

import sys
from django.conf import settings
from django.utils.importlib import import_module

from south.management.commands.migrate import Command as MigrateCommand, \
    show_migration_changes, format_migration_list_item
from south import migration
from south.migration import Migrations
from south.exceptions import NoMigrations
from south.db import DEFAULT_DB_ALIAS

__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.com'
__docformat__ = 'restructuredtext'


special_way = {
    'datalogging': getattr(settings, 'DATALOGGER_DATABASE', None),
}


class Command(MigrateCommand):
    u"""
    Команда является идентичной к South за тем исключением,
    что приложение datalogging обрабатывается отдельно.
    """

    def handle(self, app=None, target=None, skip=False, merge=False,
               backwards=False, fake=False, db_dry_run=False, show_list=False,
               show_changes=False, database=DEFAULT_DB_ALIAS,
               delete_ghosts=False, ignore_ghosts=False, **options):

        # NOTE: THIS IS DUPLICATED FROM django.core.management.commands.syncdb
        # This code imports any module named 'management' in INSTALLED_APPS.
        # The 'management' module is the preferred way of listening to post_syncdb
        # signals, and since we're sending those out with create_table migrations,
        # we need apps to behave correctly.
        for app_name in settings.INSTALLED_APPS:
            try:
                import_module('.management', app_name)
            except ImportError as exc:
                msg = exc.args[0]
                if not msg.startswith('No module named') or 'management' not in msg:
                    raise
            # END DJANGO DUPE CODE

        # if all_apps flag is set, shift app over to target
        if options.get('all_apps', False):
            target = app
            app = None

        # Migrate each app
        if app:
            try:
                apps = [Migrations(app)]
            except NoMigrations:
                print("The app '%s' does not appear to use migrations." % app)
                print("./manage.py migrate " + self.args)
                return
        else:
            apps = list(migration.all_migrations())

        # Do we need to show the list of migrations?
        if show_list and apps:
            list_migrations(apps, database, **options)

        if show_changes and apps:
            show_migration_changes(apps)

        if not (show_list or show_changes):

            for app in apps:
                _database = special_way.get(app.app_label(), database)
                if _database not in settings.DATABASES:
                    continue

                result = migration.migrate_app(
                    app,
                    target_name = target,
                    fake = fake,
                    db_dry_run = db_dry_run,
                    verbosity = int(options.get('verbosity', 0)),
                    interactive = options.get('interactive', True),
                    load_initial_data = not options.get('no_initial_data', False),
                    merge = merge,
                    skip = skip,
                    database = _database,
                    delete_ghosts = delete_ghosts,
                    ignore_ghosts = ignore_ghosts,)

                if result is False:
                    sys.exit(1) # Migration failed, so the command fails.


def list_migrations(apps, database=DEFAULT_DB_ALIAS, **options):
    """
    Prints a list of all available migrations (разбитых по БД), and which
    ones are currently applied.
    Accepts a list of Migrations instances.
    """
    from south.models import MigrationHistory
    applied_migrations = MigrationHistory.objects.filter(
        app_name__in=[app.app_label() for app in apps])

    databases = [database]
    for app in apps:
        _database = special_way.get(app.app_label())
        if not (_database is None or _database in databases):
            databases.append(_database)

    applied_migrations_lookup = {}
    for database in databases:
        applied_migrations = applied_migrations.using(database)
        key = '[{0}] {{0}}.{{1}}'.format(database)
        for mi in applied_migrations:
            mig_key = key.format(mi.app_name, mi.migration)
            applied_migrations_lookup[mig_key] = mi
    print()

    for app in apps:
        print(" " + app.app_label())
        name_tpl = '[{0}] {1}.{2}'
        # Get the migrations object
        for migration in app:
            app_label = migration.app_label()
            _database =  special_way.get(app_label)
            full_name = name_tpl.format(
                _database, app_label, migration.name())

            if full_name in applied_migrations_lookup:
                applied_migration = applied_migrations_lookup[full_name]

                print(format_migration_list_item(
                    migration.name(),
                    applied=applied_migration.applied,
                    **options))
            else:
                print(format_migration_list_item(
                    migration.name(),
                    applied=False,
                    **options))
        print()



# coding: utf-8
from django.conf import settings

__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.com'
__docformat__ = 'restructuredtext'


class DataLoggerRouter(object):
    u""" Роутер не позволяет создавать связи с моделью лога.
    """

    def db_for_read(self, model, **hints):
        u""" Выбор БД для чтения. В данном методе происходит проверка
        на заданное имя модели и в случаем если модель принадлежит
        данному модулю, для нее используются отдельные настройки
        """
        model_name = getattr(model, '__name__', model.__class__.__name__)
        if model_name == 'DataLog':
            return settings.DATALOGGER_DATABASE
        return None

    def db_for_write(self, model, **hints):
        u""" Выбор БД для записи. Применяются те же правила, что и при
        чтения из базы.
        """
        return self.db_for_read(model, **hints)

    def allow_relation(self, obj1, obj2, **hints):
        u""" Метод контролируюий факт создания связи между обектами. Отсекаются
        связи создаваемые с моделяи данного модуля.
        """
        if 'DataLog' in [obj1.__class__.__name__, obj2.__class__.__name__]:
            return False
        return None

    def allow_syncdb(self, db, model):
        u""" Метод регулирующий возможность инициализации моделей.
        """
        model_name = getattr(model, '__name__', model.__class__.__name__)
        if db == settings.DATALOGGER_DATABASE and model_name == 'DataLog':
            return True
        return None


class NotUseDataLoggerDBRouter(object):
    u""" Роутер запрещает создание таблиц в базе для логгера.
    """

    def allow_syncdb(self, db, model):
        u""" Запрет создания таблиц в базе логгера
        """
        if db == settings.DATALOGGER_DATABASE:
            if model._meta.app_label == 'south':
                return True
            return False



# coding: utf-8
# сeated on 26 дек. 2014 г.
# author: Michael Vorotyntsev

# Тут реализованы инструменты для того чтобы дамп таблицы записать
# в пустую базу средствами ORM. Если дамп таблицы большой его можно
# разбить на части по дням и записывать в несколько процессов.

import codecs
import dateutil.parser
import logging
import json
import gc
import os.path
import re
import time
from datalogging.models import DataLog
from json_field.fields import JSONDecoder
from multiprocessing import Process, Queue
from Queue import Empty


def separation_day(file_path, dir_path, sep_char=u'\t', logger_name='console'):
    """
    Распиливание дампа таблицы на части по дням.
    Вы можете сделать cvs-файл через COPY в Postgres.
    :param file_path: путь к дампу
    :param dir_path: директория где будут созданы файлы по дням
    :param sep_char: символ-разделитель полей в дампе таблицы
    :param logger_name: имя логера, для вывода информации по ходу выполнения
    """

    file_name, file_ext = os.path.basename(file_path).split('.')

    logger = logging.getLogger(logger_name)
    logger.info('конвертация...')
    start_time = time.time()
    log_step = 100000
    index = 0
    line_count = 0
    day_files = {}

    with codecs.open(file_path, 'r', encoding='utf-8') as result_file:
        line = result_file.readline()
        while line:
            line_count += 1
            index += 1
            if index >= log_step:
                index = 0
                logger.info(
                    u' обработано {} за {} дней'.format(
                        line_count, len(day_files)))
                gc.collect()

            row_data = line.split(sep_char, 3)
            # немного сжать
            row_data[3] = re.sub(
                r'\\\\', r'\\', re.sub(r'\\N', r'', row_data[3]))

            record_date = dateutil.parser.parse(row_data[2]).date()
            if record_date not in day_files:
                out_file = codecs.open(
                    os.path.join(
                        dir_path,
                        '{}_{}.{}'.format(
                            file_name,
                            record_date.isoformat().replace('-', '_'),
                            file_ext)),
                    mode='w',
                    encoding='utf-8')
                day_files[record_date] = out_file
            else:
                out_file = day_files[record_date]

            new_line = sep_char.join(row_data)
            out_file.write(new_line)

            line = result_file.readline()

        # закрыть файлы
        for done_file in day_files.values():
            done_file.close()

    logger.info((
        u'Время выполнения: {:.2f} мин. фйлов'
        u'создано {} записей обработано {}').format(
            (time.time() - start_time) / 60.0,
            len(day_files),
            line_count))


def _fill_db_worker(
        file_path,
        file_name,
        worker_num,
        free_workers,
        safe=True,
        remove_done=False,
        sep_char='\t',
        logger_name='console'):
    """
    Процесс работающий с файлом-источником записей для таблицы.
    """

    def convert_json(data):
        return json.loads(data, cls=JSONDecoder)

    fields = [
        (0, 'id', long),
        (1, 'guid', None),
        (2, 'timestamp', dateutil.parser.parse),
        (3, 'event_type', None),
        (4, 'event_code', None),
        (5, 'object_id', long),
        (6, 'object_type', None),
        (7, 'object_snapshot', convert_json),
        (8, 'verbose', None),
        (9, 'system_type', None),
        (10, 'suid', long),
        (11, 'ip', None),
        (12, 'context_data', convert_json),
        (13, 'request_token', None),
    ]

    logger = logging.getLogger(logger_name)
    start_time = time.time()

    log_step = 50000
    created = 0
    i = 0
    insert_time = 0.001

    with codecs.open(file_path, 'r', encoding='utf-8') as result_file:
        line = result_file.readline()
        while line:
            created += 1
            i += 1
            if i > log_step:
                i = 0
                logger.info(
                    u"{}: обработано: {}".format(file_name, created))
                gc.collect()

            row_date = line.split(sep_char)
            new_record = DataLog()
            for index, field, convertor in fields:
                data = row_date[index]
                if data:
                    if convertor:
                        try:
                            data = convertor(data)
                        except (ValueError, TypeError):
                            data = None
                else:
                    data = None

                if data:
                    setattr(new_record, field, data)

            created += 1
            insert_time_start = time.time()
            if not safe:
                new_record.save()
            insert_time = (
                insert_time + (time.time() - insert_time_start)) / 2.0

            line = result_file.readline()

    logger.info(
        u'{}: всего записей создано: {} за {:.2f} мин.'.format(
            file_name, created, (time.time() - start_time) / 60.0))
    logger.info(
        u'{} среднее время записи: {:.12f}'.format(
            file_name, insert_time))

    # отправить информацию о завершении
    free_workers.put(worker_num)
    # удалить обработанный файл
    if not safe and remove_done:
        os.remove(file_path)


def multiworker_fill(
        day_files_path,
        pool_size=12,
        safe=True,
        remove_done=False,
        sep_char='\t',
        logger_name='console'):
    """
    Запись данных в таблицу логов в несколько процессов из файлов данных
    разбитых по индексам или по дням.

    :param day_files_path: директория с файлами
    :param pool_size: количество процессов (размер пула)
    :param safe: безопасный режим (в базу не пишется, файлы не удаляются)
    :param remove_done: удалять обработанные файлы
        (полезно если у вас мало места на диске)
    :param sep_char: символ-разделитель полей в дампе таблицы
    :param logger_name: имя логера, для вывода информации по ходу выполнения
    """
    workers = {}
    start_time = time.time()
    for index in xrange(pool_size):
        workers[index + 1] = dict(
            name=None, pid=None)

    files = {}
    for name in os.listdir(day_files_path):
        file_path = os.path.join(day_files_path, name)
        if os.path.isfile(file_path):
            files[name] = {
                'path': file_path,
                'done': False,
                'run': False,
            }

    logger = logging.getLogger(logger_name)
    if not files:
        raise Exception()

    logger.info(u"Найдено {} файлов".format(len(files)))

    iter_delay = 2.0
    free_workers = Queue()
    file_done_count = 0

    def file_not_done(_file_data):
        return not _file_data[1].get('done')

    while any(map(file_not_done, files.iteritems())):
        # если есть свободный воркер его надо занять
        try:
            now_free = free_workers.get(block=False)
        except Empty:
            now_free = None
        else:
            logger.info(' >> освободился: {}'.format(now_free))

        for worker_num, worke_data in workers.iteritems():
            file_name = worke_data.get('name')
            pid = worke_data.get('pid')

            if pid:
                # через is_alive и вообще через pid
                # не получается, залипает зомби из-за использования
                # stdout (и stderr) в главном и дочерних процессах
                # обойти через .communicate() не получилось
                # зомби-pid-ы иногда сами подыхают но не сразу, как должно быть
                # если родительский процесс игнорирует SIGCHLD (
                # а он игнорируется по умолчанию),
                # то зомби остаются до завершения родительского процесса
                is_free = now_free == worker_num

                if is_free:
                    # завершился!
                    files[file_name].update(
                        run=False, done=True)
                    worke_data.update(pid=None, name=None)
                    file_done_count += 1
                    logger.info(
                        u' файлов выполнено {}'.format(file_done_count))
            else:
                is_free = True

            # можно запустить обработку для
            if is_free:

                # найти свободный файл не выполненный
                file_name = None
                for new_file_name, file_db_data in files.iteritems():
                    if file_db_data['done'] or file_db_data['run']:
                        # готов или обрабатывается
                        continue
                    # есть
                    file_name = new_file_name
                    break

                if file_name:
                    file_path = files[file_name].get('path')
                    proc = Process(
                        target=_fill_db_worker,
                        kwargs={
                            'worker_num': worker_num,
                            'free_workers': free_workers,
                            'file_name': file_name,
                            'file_path': file_path,
                            'remove_done': remove_done,
                            'safe': safe,
                            'logger_name': logger_name,
                            'sep_char': sep_char,
                        })

                    files[file_name].update(run=True, done=False)
                    proc.start()

                    worke_data.update(pid=proc.pid, name=file_name)
                    logger.info((
                        u'Будет запущен обработчик {} PID:{} '
                        u'для файла {}').format(
                            worker_num, proc.pid, file_path))
        # задержка перед каждой попыткой найти завершенный воркер
        # и создать новый, новый содается пока весь пул не будет
        # занят
        time.sleep(iter_delay)

    logger.info(
        u"Время выполнения: {0:.2f} мин.".format(
            (time.time() - start_time) / 60.0))



# coding: utf-8
import uuid
from django.db import models
from django.db.models.signals import class_prepared
from m3_django_compat import get_model

from .helpers import get_snapshot

__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.com'
__docformat__ = 'restructuredtext'


MANAGER_NAME = '_datalogger_manager'


class DataLoggingManager(object):

    #: Серия UID для обработчиков сигналов, устанавливается в качестве
    #: значения для ``dispatch_uid``. Дополнительную информацию о
    #: ``dispatch_uid`` см. ``django.db.models.signals.Signal.connect``.
    POST_INIT_UID = uuid.uuid4().hex
    POST_SAVE_UID = uuid.uuid4().hex
    POST_DELETE_UID = uuid.uuid4().hex

    def contribute_to_class(self, cls, name):
        u""" Привязка к модели.

        Если `name` является DATALOGGER_MANAGER_NAME, то привязка
        к модели добавлена через settings.py и сигнал class_prepared
        уже возник, соответственно вешать на него `finalize`
        бессмысленно.

        """
        if name != 'DataLog':
            if name == MANAGER_NAME:
                self.finalize(cls)
            else:
                class_prepared.connect(self.finalize)
            setattr(cls, name, True)

    def finalize(self, sender, **kw):
        u""" Подключение к модели. """
        models.signals.post_init.connect(
            self.post_init, sender=sender, weak=False,
            dispatch_uid=self.POST_INIT_UID)
        models.signals.post_save.connect(
            self.post_save, sender=sender, weak=False,
            dispatch_uid=self.POST_SAVE_UID)
        models.signals.post_delete.connect(
            self.post_delete, sender=sender, weak=False,
            dispatch_uid=self.POST_DELETE_UID)

    def post_init(self, instance, **kw):
        u""" Сохранение начального состояния модели. """
        state = get_snapshot(instance)
        instance._snapshot = state

    def post_save(self, instance, created, **kw):
        u""" Сохранение состояния модели в лог при сохранении. """
        from .enums import EventCode

        DataLog = get_model('datalogging', 'DataLog')
        changed = get_snapshot(instance)
        if changed != instance._snapshot:
            event_code = created and EventCode.INSERT or EventCode.UPDATE
            DataLog.make_snapshot(instance, event_code)
            instance._snapshot = get_snapshot(instance)

    def post_delete(self, instance, **kw):
        u""" Сохранение состояния модели в лог перед удалением. """
        from .enums import EventCode

        DataLog = get_model('datalogging', 'DataLog')
        DataLog.make_snapshot(instance, EventCode.DELETE)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'ParanoidLog'
        db.create_table(u'data_logging', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('guid', self.gf('django.db.models.fields.CharField')(max_length=64)),
            ('timestamp', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('event_type', self.gf('django.db.models.fields.CharField')(max_length=256)),
            ('event_code', self.gf('django.db.models.fields.CharField')(max_length=32)),
            ('object_id', self.gf('django.db.models.fields.IntegerField')(null=True)),
            ('object_type', self.gf('django.db.models.fields.CharField')(max_length=96, null=True)),
            ('object_snapshot', self.gf('django.db.models.fields.TextField')(default='{}', null=True)),
            ('verbose', self.gf('django.db.models.fields.TextField')()),
            ('system_type', self.gf('django.db.models.fields.CharField')(max_length=32)),
            ('suid', self.gf('django.db.models.fields.IntegerField')(null=True)),
            ('ip', self.gf('django.db.models.fields.CharField')(max_length=15)),
            ('context_data', self.gf('django.db.models.fields.TextField')(default='{}', null=True)),
            ('request_token', self.gf('django.db.models.fields.CharField')(max_length=64, db_index=True)),
        ))
        db.send_create_signal('datalogging', ['DataLog'])


    def backwards(self, orm):
        # Deleting model 'ParanoidLog'
        db.delete_table(u'm3_paranoid_log')


    models = {
        'datalogging.datalog': {
            'Meta': {'object_name': 'DataLog', 'db_table': "u'data_logging'"},
            'context_data': ('django.db.models.fields.TextField', [], {'default': "'{}'", 'null': 'True'}),
            'event_code': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'event_type': ('django.db.models.fields.CharField', [], {'max_length': '256'}),
            'guid': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'ip': ('django.db.models.fields.CharField', [], {'max_length': '15'}),
            'object_id': ('django.db.models.fields.IntegerField', [], {'null': 'True'}),
            'object_snapshot': ('django.db.models.fields.TextField', [], {'default': "'{}'", 'null': 'True'}),
            'object_type': ('django.db.models.fields.CharField', [], {'max_length': '96', 'null': 'True'}),
            'request_token': ('django.db.models.fields.CharField', [], {'max_length': '64', 'db_index': 'True'}),
            'suid': ('django.db.models.fields.IntegerField', [], {'null': 'True'}),
            'system_type': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'timestamp': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'verbose': ('django.db.models.fields.TextField', [], {})
        }
    }

    complete_apps = ['datalogging']



# coding: utf-8
from django.conf import settings

__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.com'
__docformat__ = 'restructuredtext'

from django.contrib.auth.models import AnonymousUser

from m3.actions import ActionPack, ActionController

from .helpers import remember_user, memorize_user
from .models import DataLog
from .enums import EventCode

EXCLUDED_ACTIONS = getattr(settings, 'DATALOGGER_EXCLUDE_ACTIONS', ())
DATALOGGER_OFF = getattr(settings, 'DATALOGGER_SHUTUP', False)
FORGET_SYS_EVENTS = getattr(settings, 'DATALOGGER_FORGET_SYS_EVENTS', False)
HOOKED_ACTIONS = getattr(settings, 'DATALOGGER_HOOKED_ACTIONS', {})

orig_invoke = ActionController._invoke


class InvaderLogPack(ActionPack):
    u""" Инжектируемый пак
    """

    def pre_run(self, request, context):
        u""" Перед выполнением экшена у инжектируемого пака
        необходимо запоминить пользователя
        """
        memorize_user(request.user)

    def post_run(self, request, context, response):
        u""" Некоторые косвенные методы определения поведения пользователя.
        Если в ответе на запрос упоминается заголовок, то считается, что
        вернули окно.
        Если запомненный ранее пользователь был аутентифицирован, а после
        формирования ответа - нет, то считается, что произошел выход из
        системы. И наоборот.
        """
        user = remember_user()
        if hasattr(response, 'data') and hasattr(response.data, 'title'):
            win = response.data
            DataLog.make_system_event(EventCode.WIN_OPEN, {
                'title': win.title
            })
        elif isinstance(request.user, AnonymousUser) and user != request.user:
            DataLog.make_system_event(EventCode.LOGOUT)
        elif isinstance(user, AnonymousUser) and user != request.user:
            DataLog.make_system_event(EventCode.LOGIN)


# Подмена оригинального метода, на метод который осуществляет инъекцию
# в InvaderLogPack в stack.

_invader = InvaderLogPack()


def is_loggable(obj):
    u"""Данная функция использоуется для опредения нуждается ли данный объект
    в логировании. Проводится проверка объекта на вхождение имени его класса,
    либо имени класса его предка в список-исключение `EXCLUDED_ACTIONS`. Если
    переданный экземпляр найден в списке, то объект счтается подходящим для
    создания записи.

    :param obj: Экземпляр Action'a.
    :type obj: Action
    :returns:  bool -- результат проверки.

    """
    return not (obj.__class__.__name__ in EXCLUDED_ACTIONS or
                obj.parent.__class__.__name__ in EXCLUDED_ACTIONS)


def invader_invoke(self, request, action, stack, *args, **kwargs):
    u"""
    :param request: Экземпляр Action'a.
    :type request: request
    :param action: Экземпляр Action'a.
    :type action: Action
    :param stack: Экземпляр Action'a.
    :type stack: list
    """
    if is_loggable(action) and _invader not in stack:
        stack.append(_invader)

    if action.__class__.__name__ in HOOKED_ACTIONS:
        event_key = HOOKED_ACTIONS[action.__class__.__name__]
        event_code = getattr(EventCode, event_key)
        DataLog.make_system_event(event_code, {}, action=action,
            request=request)

    return orig_invoke(self, request, action, stack, *args, **kwargs)

# Проверка на частичное или полное отключение логирования
if not (DATALOGGER_OFF or FORGET_SYS_EVENTS):
    ActionController._invoke = invader_invoke



# coding: utf-8
from django.conf import settings
from django.utils.encoding import force_str

from m3_legacy.middleware import get_thread_data

__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.com'
__docformat__ = 'restructuredtext'


def get_snapshot(instance):
    u""" Получение снимка состояния экземпляра модели """
    state = set()
    for attr_name, val in instance.__dict__.items():
        if attr_name.startswith('_') or attr_name in settings.DATALOGGER_EXCLUDE_FIELDS:
            continue

        if val is None:
            state.add((attr_name, val))
        else:
            str_val = force_str(val)
            if str_val.isdigit():
                state.add((attr_name, val))
            else:
                state.add((attr_name, str_val))

    return state


def is_server_mode():
    u""" Определение режима работы сервера.
    """
    _thread_locals = get_thread_data()
    return hasattr(_thread_locals, 'request_token')


def only_server_mode(func):
    u""" Предотвращение запуска логгера в шеле
    """
    def wrapper(*a, **kw):
        if is_server_mode():
            return func(*a, **kw)
        else:
            return None
    return wrapper


def memorize_user(user):
    u""" Запоминание пользователя в текущем треде.
    """
    _thread_locals = get_thread_data()
    _thread_locals.suspicious_user = user


def remember_user():
    u""" Вспоминаем пользователя
    """
    _thread_locals = get_thread_data()
    return getattr(_thread_locals, 'suspicious_user', None)



# coding: utf-8
"""Логгер действий пользователя в системе.

.. moduleauthor:: Andrey Torsunov <torsunov@bars-open.ru>

"""

__author__ = 'Andrey Torsunov'
__contact__ = 'torsunov@bars-open.ru'
__docformat__ = 'restructuredtext'



# coding: utf-8
from setuptools import setup, find_packages


setup(
    name='m3-datalogging',
    version='2.0.3.13',
    author='Andrew Torsunov, Kirill Borisov',
    author_email='torsunov@bars-open.ru, borisov@bars-open.ru',
    packages=find_packages('src', exclude=['tests']),
    package_dir={'': 'src'},
    include_package_data=True,
    url='https://bitbucket.org/barsgroup/data-logging',
    license='MIT',
    description='Logging system to spy on users.',
    exclude_package_data={'': ['*.pyc', '*.pyo']},
    classifiers=[
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Natural Language :: Russian',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 5 - Production/Stable',
    ],
    install_requires=[
        'django-json-field>=0.5.8',
        'm3_legacy>=2.1.1',
        'architect==0.5.0',
    ],
)
